function  [uroot, planinfo] = opd(xroot, model, cfg)
% OPD: Optimstic planning for deterministic systems
% Author: Lucian Busoniu

CFG.discount = [];              % discount factor
CFG.n = [];                     % expansion budget
CFG.U = [];                     % discrete, flat action space

cfg = checkparams(cfg, CFG);

% quick-access to config parameters
% (note to avoid conflict with 'gamma' function, we must name discount factor differently)
discount = cfg.discount; n = cfg.n;
U = cfg.U;
M = size(U, 2);      % number of discrete actions
dimx = model.p;      % number of state vars

t = cputime;        % start counting execution time

maxtreesize = M*n + 1;
% prepare tree data 
Tparent = zeros(2, maxtreesize, 'uint32');        % parent and parent action
Tchildren = zeros(M, maxtreesize, 'uint32');      % child index for each action
Tdepth = zeros(1, maxtreesize, 'uint16');         % depth (uint16 as depth > 65536 considered unfeasible)
Tleaf = false(1, maxtreesize);                    % leaf flag
Tx = NaN(dimx, maxtreesize, 'double');            % state
Tell = NaN(1, maxtreesize, 'double');             % partial return up to node (cumulative reward)
Tb = NaN(1, maxtreesize, 'double');

% initialize the root node (note 0 = null pointer)
root = 1;
Tparent(:, root) = 0;
Tchildren(:, root) = 0;
Tleaf(root) = true;
Tdepth(root) = 0;
Tx(:, root) = xroot;
Tell(root) = 0;
Tb(root) = 1 / (1-discount);
treesize = 1;

% main tree construction loop
for tau = 1:n,
    
    % select candidate leaves for expansion
    % uniform selection: smallest depth
    leaves = find(Tleaf);
    [bmax, opt] = max(Tb(leaves));
    % break ties in favor of the first candidate
    nodetoexpand = leaves(opt(1));
        
    % expand the chosen leaf
    Tleaf(nodetoexpand) = false;        % no longer a leaf
    % run through the discrete actions 
    for j = 1:M,
        % find index where new node will be, save it on the children table of current node
        child = treesize + 1;
        Tchildren(j, nodetoexpand) = child;
        
        % simulate transition and add data to tree
        [xplus, reward] = feval(model.fun, model, Tx(:, nodetoexpand), U(:, j));

        % create new node
        Tparent(:, child) = [nodetoexpand; j];      % parent is expanded node, along jth action
        Tchildren(:, child) = 0;                    % init children pointers to "null"
        Tdepth(child) = Tdepth(nodetoexpand) + 1;   % one deeper than the parent
        Tleaf(child) = true;                        % new child is a leaf
        Tx(:, child) = xplus;                       % state of the child
        % partial return is the parent's return + the (appropriately discounted) reward
        % note the conversion of the depth to from int to double 
        Tell(child) = Tell(nodetoexpand) + discount^double(Tdepth(nodetoexpand)) * reward;
        Tb(child) = Tell(child) + discount^double(Tdepth(child)) / (1-discount);
            
        % we just added a new node, increase tree size
        treesize = treesize + 1;
    end;
    
end;    % end main tree construction loop


% find action at the tree root
% choose a "greedy" leaf with maximal ell-value (truncated return)
leaves = find(Tleaf); [ellmax, jmax] = max(Tell(leaves)); greedy = leaves(jmax(1)); 
% navigate greedy path upwards in the tree until level 1, just after root
while Tparent(1, greedy) ~= root, greedy = Tparent(1, greedy); end;
% choose the action that was taken at root to arrive at this level-1 node
% -- this is the first action on the sequence
uroot = U(:, Tparent(2, greedy));

% output stats
planinfo.n = tau;
planinfo.depth = max(Tdepth);
planinfo.trun = cputime - t;
planinfo.treesize = treesize;
planinfo.message = sprintf('Planning done, n=%d expansions, depth=%d, treesize=%d', ...
    tau, planinfo.depth, planinfo.treesize); 

end % main function

